/*
 *  Copyright (C) 2002-2005 The SSI Project. All rights reserved.
 *  http://ssi.is.s.u-tokyo.ac.jp
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY.  See the COPYING file for the precise
 *  terms and conditions for the use of this software.
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "fftss.h"  
#include <math.h>

extern void fftss_check_cpu();

int main(int argc, char **argv)
{
  long i, j;
  double *b0, *b1;
  long logn = 10;
  long n = 1 << 10;
  fftss_plan p0, p1;
  double t0, t1, t2 = 0.0;
  long global_tries = 10;
  long tries = 30;

  fftss_check_cpu();

  if (argc > 1) logn = atol(argv[1]);
  tries = 300 / logn;
  n = 1 << logn;

  b0 = fftss_malloc(sizeof(double) * 2 * n);
  b1 = fftss_malloc(sizeof(double) * 2 * n);
  
  fftss_test_dft_1d(n, b0, b1, FFTSS_BACKWARD, 0);
  fftss_test_dft_1d(n, b1, b0, FFTSS_FORWARD, 0);

#if 1
  p0 = fftss_plan_dft_1d(n, b0, b1, FFTSS_BACKWARD, FFTSS_VERBOSE);
  p1 = fftss_plan_dft_1d(n, b1, b0, FFTSS_FORWARD, FFTSS_VERBOSE);
#else
  p0 = fftss_plan_dft_1d(n, b0, b1, FFTSS_BACKWARD, FFTSS_VERBOSE|FFTSS_ESTIMATE);
  p1 = fftss_plan_dft_1d(n, b1, b0, FFTSS_FORWARD, FFTSS_VERBOSE|FFTSS_ESTIMATE);
#endif

  for (j = 0; j < global_tries; j++) {

  for (i = 0; i < n; i++) {
#if 1
    b0[2 * i] = (double)i + 1.0;
    b0[2 * i + 1] = (double)(n - i);
#else
    b0[2 * i] = 1.0;
    b0[2 * i + 1] = 0.0;
#endif
  }
#if 0
  b0[2] = 1.0;
#endif
  for (i = 0; i < 2 * n; i++) b1[i] = 0.0;

  t0 = fftss_get_wtime();
  for (i = 0; i < tries; i++) {
    fftss_execute(p0);
    fftss_execute(p1);
  }
  t1 = fftss_get_wtime() - t0;
  t2 += t1;

  t0 = 5.0 * (double)n * log((double)n) / log(2.0) * 2.0 * (double)tries;
  
  printf("[%02ld]%lf sec. %lf Mflops.\n", j,
	 t1 / (double)tries, t0 / t1 * 0.000001);
  }

  t0 = 5.0 * (double)n * log((double)n) / log(2.0) * 2.0 * (double)tries;
  t0 *= (double)global_tries;
  printf("= %lf sec. %lf Mflops.\n",
         t2 / (double)tries, t0 / t2 * 0.000001);
 
  t0 = (double)1.0 / (double)n;
  t1 = 1.0;
  for (i = 0; i < tries; i++) t1 *= t0;


  for (i = 0; i < 2 * n; i++) b0[i] *= t1;
  
#if 0
  for (i = 0; i < n; i++)
    printf("[%ld] %le,%le\n", i, b0[2 * i], b0[2 * i + 1]);
#endif

  for (i = 0; i < n; i++) {
    b0[2 * i] -= (double)i + 1.0;
    b0[2 * i + 1] -= (double)(n - i);
  }
  for (i = 0; i < 2 * n; i++) b0[i] = fabs(b0[i]);

  {
    double max_err = 0.0;

    for (i = 0; i < n; i++) 
      if (b0[2 * i] > max_err) max_err = b0[2 * i];
    printf("Max error=%le\n", max_err);
  }
}
